"""Wobbly Vapoursynth Functions."""

__version__ = '0.2.2'

__author__ = 'LightArrowsEXE <LightArrowsReboot@gmail.com>'
__maintainer__ = __author__

__author_name__, __author_email__ = [x[:-1] for x in __author__.split('<')]
__maintainer_name__, __maintainer_email__ = [x[:-1] for x in __maintainer__.split('<')]

if __name__ == '__github__':
    print(__version__)
